
.. DO NOT EDIT.
.. THIS FILE WAS AUTOMATICALLY GENERATED BY SPHINX-GALLERY.
.. TO MAKE CHANGES, EDIT THE SOURCE PYTHON FILE:
.. "auto_examples/anatomical/example_T2s_EPI.py"
.. LINE NUMBERS ARE GIVEN BELOW.

.. only:: html

    .. note::
        :class: sphx-glr-download-link-note

        :ref:`Go to the end <sphx_glr_download_auto_examples_anatomical_example_T2s_EPI.py>`
        to download the full example code.

.. rst-class:: sphx-glr-example-title

.. _sphx_glr_auto_examples_anatomical_example_T2s_EPI.py:


Compare Fourier Model and T2* Model for EPI
===========================================

This examples walks through the elementary components of SNAKE.

Here we proceed step by step and use the Python interface. A more integrated
alternative is to use the CLI ``snake-main``

.. GENERATED FROM PYTHON SOURCE LINES 14-21

.. code-block:: Python


    # Imports
    import numpy as np
    from snake.core.simulation import SimConfig, default_hardware, GreConfig
    from snake.core.phantom import Phantom
    from snake.core.sampling import EPI3dAcquisitionSampler








.. GENERATED FROM PYTHON SOURCE LINES 22-36

.. code-block:: Python


    sim_conf = SimConfig(
        max_sim_time=6,
        seq=GreConfig(TR=100, TE=30, FA=3),
        hardware=default_hardware,
    )
    sim_conf.hardware.n_coils = 8
    sim_conf.fov.res_mm = (3, 3, 3)

    phantom = Phantom.from_brainweb(
        sub_id=4, sim_conf=sim_conf, tissue_file="tissue_7T", output_res=1
    )






.. rst-class:: sphx-glr-script-out

 .. code-block:: none

    No matrix size found in the header. The header is probably missing.




.. GENERATED FROM PYTHON SOURCE LINES 37-39

Setting up Acquisition Pattern and Initializing Result file.
------------------------------------------------------------

.. GENERATED FROM PYTHON SOURCE LINES 39-47

.. code-block:: Python


    # The next piece of simulation is the acquisition trajectory.
    # Here nothing fancy, we are using a EPI (fully sampled), that samples a 3D
    # k-space (this akin to the 3D EPI sequence of XXXX)

    sampler = EPI3dAcquisitionSampler(accelz=1, acsz=0.1, orderz="top-down")









.. GENERATED FROM PYTHON SOURCE LINES 48-62

Acquisition with Cartesian Engine
---------------------------------

The generated file ``example_EPI.mrd`` does not contains any k-space data for
now, only the sampling trajectory. let's put some in. In order to do so, we
need to setup the **acquisition engine** that models the MR physics, and get
sampled at the specified k-space trajectory.

SNAKE comes with two models for the MR Physics:

- model="simple" :: Each k-space shot acquires a constant signal, which is the
  image contrast at TE.
- model="T2s" :: Each k-space shot is degraded by the T2* decay induced by
  each tissue.

.. GENERATED FROM PYTHON SOURCE LINES 62-97

.. code-block:: Python


    # Here we will use the "simple" model, which is faster.
    #
    # SNAKE's Engine are capable of simulating the data in parallel, by distributing
    # the shots to be acquired to a set of processes. To do so , we need to specify
    # the number of jobs that will run in parallel, as well as the size of a job.
    # Setting the job size and the number of jobs can have a great impact on total
    # runtime and memory consumption.
    #
    # Here, we have a single frame to acquire with 60 frames (one EPI per slice), so
    # a single worker will do.

    from snake.core.engine import EPIAcquisitionEngine

    engine = EPIAcquisitionEngine(model="simple")

    engine(
        "example_EPI.mrd",
        sampler,
        phantom,
        sim_conf,
        worker_chunk_size=60,
        n_workers=1,
    )
    engine_t2s = EPIAcquisitionEngine(model="T2s")

    engine_t2s(
        "example_EPI_t2s.mrd",
        sampler,
        phantom,
        sim_conf,
        worker_chunk_size=60,
        n_workers=1,
    )



.. rst-class:: sphx-glr-script-out

.. code-block:: pytb

    Traceback (most recent call last):
      File "/volatile/github-ci-mind-inria/gpu_mind_runner/_work/snake-fmri/snake-fmri/examples/anatomical/example_T2s_EPI.py", line 78, in <module>
        engine(
      File "/volatile/github-ci-mind-inria/gpu_mind_runner/_work/snake-fmri/snake-fmri/src/snake/core/engine/base.py", line 299, in __call__
        raise exc
      File "/volatile/github-ci-mind-inria/gpu_mind_runner/_work/snake-fmri/snake-fmri/src/snake/core/engine/base.py", line 296, in __call__
        f_chunk = str(future.result())
      File "/volatile/github-ci-mind-inria/gpu_mind_runner/_work/_tool/Python/3.10.16/x64/lib/python3.10/concurrent/futures/_base.py", line 451, in result
        return self.__get_result()
      File "/volatile/github-ci-mind-inria/gpu_mind_runner/_work/_tool/Python/3.10.16/x64/lib/python3.10/concurrent/futures/_base.py", line 403, in __get_result
        raise self._exception
      File "/volatile/github-ci-mind-inria/gpu_mind_runner/_work/_tool/Python/3.10.16/x64/lib/python3.10/concurrent/futures/thread.py", line 58, in run
        result = self.fn(*self.args, **self.kwargs)
      File "/volatile/github-ci-mind-inria/gpu_mind_runner/_work/snake-fmri/snake-fmri/src/snake/core/engine/base.py", line 161, in _acquire_ksp_job
        ksp = _job_model(phantom, ddatas, sim_conf, trajs, **kwargs)
      File "/volatile/github-ci-mind-inria/gpu_mind_runner/_work/snake-fmri/snake-fmri/src/snake/core/engine/cartesian.py", line 169, in _job_model_simple
        ksp_coil = ksp[c]
    IndexError: index 1 is out of bounds for axis 0 with size 1




.. GENERATED FROM PYTHON SOURCE LINES 98-110

Simple reconstruction
---------------------

Getting k-space data is nice, but
SNAKE also provides rudimentary reconstruction tools to get images (and check
that we didn't mess up the acquisition process).
This is available in the companion package ``snake.toolkit``.

Loading the ``.mrd`` file to retrieve all information can be done using the
``ismrmd`` python package, but SNAKE provides convenient dataloaders, which are
more efficient, and take cares of managing underlying files access. As we are
showcasing the API, we will do things manually here, and use only core SNAKE.

.. GENERATED FROM PYTHON SOURCE LINES 110-133

.. code-block:: Python


    from snake.mrd_utils import CartesianFrameDataLoader
    from scipy.fft import ifftn, ifftshift, fftshift


    def reconstruct_frame(filename):
        with CartesianFrameDataLoader(filename) as data_loader:
            _, kspace_data = data_loader.get_kspace_frame(0)
        axes = (-3, -2, -1)
        image_data = ifftshift(
            ifftn(fftshift(kspace_data, axes=axes), axes=axes, norm="ortho"), axes=axes
        )

        # Take the square root sum of squares to get the magnitude image (SSOS)
        image_data = np.sqrt(np.sum(np.abs(image_data) ** 2, axis=0))

        return image_data.squeeze().T


    image_simple = reconstruct_frame("example_EPI.mrd")
    image_T2s = reconstruct_frame("example_EPI_t2s.mrd")



.. GENERATED FROM PYTHON SOURCE LINES 134-136

Plotting the result
-------------------

.. GENERATED FROM PYTHON SOURCE LINES 136-152

.. code-block:: Python


    import matplotlib.pyplot as plt
    from snake.toolkit.plotting import axis3dcut

    fig, axs = plt.subplots(1, 3, figsize=(30, 10))

    for ax, img, title in zip(
        axs,
        (image_simple, image_T2s, abs(image_simple - image_T2s)),
        ("simple", "T2s", "diff"), strict=False,
    ):
        axis3dcut(img, None, None, cbar=True, cuts=(40, 40, 40), width_inches=4, ax=ax)
        ax.set_title(title)

    plt.show()



.. rst-class:: sphx-glr-timing

   **Total running time of the script:** (0 minutes 7.997 seconds)


.. _sphx_glr_download_auto_examples_anatomical_example_T2s_EPI.py:

.. only:: html

  .. container:: sphx-glr-footer sphx-glr-footer-example

    .. container:: sphx-glr-download sphx-glr-download-jupyter

      :download:`Download Jupyter notebook: example_T2s_EPI.ipynb <example_T2s_EPI.ipynb>`

    .. container:: sphx-glr-download sphx-glr-download-python

      :download:`Download Python source code: example_T2s_EPI.py <example_T2s_EPI.py>`

    .. container:: sphx-glr-download sphx-glr-download-zip

      :download:`Download zipped: example_T2s_EPI.zip <example_T2s_EPI.zip>`


.. only:: html

 .. rst-class:: sphx-glr-signature

    `Gallery generated by Sphinx-Gallery <https://sphinx-gallery.github.io>`_
